import asyncio
import asyncio.subprocess
import os
import shlex
import subprocess
import sys
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

import build.env
import click
import packaging.version
import tomlkit
from pkg_resources import Requirement

import build
from vulcan import Vulcan, flatten_reqs
from vulcan.build_backend import (get_pip_version, get_virtualenv_python,
                                  install_develop)
from vulcan.builder import resolve_deps

version: Callable[[str], str]
if sys.version_info >= (3, 8):
    from importlib.metadata import PackageNotFoundError, version
else:
    from importlib_metadata import PackageNotFoundError, version

pass_vulcan = click.make_pass_decorator(Vulcan)

try:
    vulcan_version = version('vulcan-py')
except PackageNotFoundError:
    vulcan_version = '0.0.0'


@click.group()
@click.version_option(vulcan_version)
@click.pass_context
def main(ctx: click.Context) -> None:
    if sys.platform == 'win32':
        # This is required for python <3.8, where this is not yet the default for windows.
        # Can be removed when support for those are removed.
        # https://docs.python.org/3.7/library/asyncio-eventloop.html#event-loop-implementations
        #
        # > By default asyncio is configured to use SelectorEventLoop on all platforms.
        #
        # And then https://docs.python.org/3.8/library/asyncio-eventloop.html#event-loop-implementations
        #
        # > By default asyncio is configured to use SelectorEventLoop on Unix and ProactorEventLoop on
        #   Windows.
        asyncio.set_event_loop(asyncio.ProactorEventLoop())
    # don't fail on missing lock here, because this config object is not actually used for building only for
    # cli values
    ctx.obj = Vulcan.from_source(Path().absolute(), fail_on_missing_lock=False)


async def build_shiv_apps(from_dist: str, vulcan: Vulcan, outdir: Path) -> List[Path]:
    results = []
    for app in vulcan.shiv_options:
        try:
            if app.with_extras:
                dist = f'{from_dist}[{",".join(app.with_extras)}]'
            else:
                dist = from_dist
            cmd = [sys.executable, '-m', 'shiv', dist, '-o', str(outdir / app.bin_name)]
            if app.console_script:
                cmd += ['-c', app.console_script]
            if app.entry_point:
                cmd += ['-e', app.entry_point]
            if app.interpreter:
                cmd += ['-p', app.interpreter]
            if app.extra_args:
                cmd += shlex.split(app.extra_args)
            proc = await asyncio.subprocess.create_subprocess_exec(*cmd)
            results.append((proc, app.bin_name))
        except KeyError as e:
            raise KeyError('missing config value in pyproject.toml: {e}') from e
    await asyncio.gather(*(p.wait() for p, _ in results))
    succeeded = []
    failed = []
    for proc, res in results:
        if await proc.wait() == 0:
            succeeded.append(outdir / res)
        else:
            failed.append(res)
    for fail in failed:
        print(f"failed to create executable {failed}", file=sys.stderr)
    return succeeded


@main.command(name='build')
@click.option('--outdir', '-o', default='dist/', type=Path)
@click.option('--lock/--no-lock', '_lock', default=True)
@click.option('--wheel', is_flag=True, default=False)
@click.option('--sdist', is_flag=True, default=False)
@click.option('--shiv', is_flag=True, default=False)
@pass_vulcan
def build_out(config: Vulcan, outdir: Path, _lock: bool, wheel: bool, sdist: bool, shiv: bool) -> None:
    "Create wheels, sdists, and shiv executables"
    # for ease of use
    if len([v for v in (shiv, wheel, sdist) if v]) != 1:
        raise click.UsageError("Must specify exactly 1 of --shiv, --wheel, or --sdist")

    should_lock = _lock and not config.no_lock

    config_settings = {}
    if not _lock:
        config_settings['no-lock'] = 'true'
    project = build.ProjectBuilder('.')
    outdir.mkdir(exist_ok=True)
    if sdist:
        dist = project.build('sdist', str(outdir), config_settings=config_settings)
    elif wheel or shiv:
        if shiv and not should_lock:
            raise click.UsageError("May not specify both --shiv and --no-lock; shiv builds must be locked")
        dist = project.build('wheel', str(outdir), config_settings=config_settings)
    else:
        assert False, 'unreachable because dist_types is required'
    if shiv:
        try:
            asyncio.get_event_loop().run_until_complete(build_shiv_apps(dist, config, outdir))
        finally:
            os.remove(dist)


async def resolve_deps_or_report(config: Vulcan, python_version: str = None) -> Tuple[List[str], Dict[str, List[str]]]:
    try:
        return await resolve_deps(flatten_reqs(config.configured_dependencies),
                                  config.configured_extras or {},
                                  python_version)

    except subprocess.CalledProcessError as e:
        print(e.stderr.decode(), file=sys.stderr)
        raise


@main.command()
@pass_vulcan
def lock(config: Vulcan) -> None:
    "Generate and update lockfile"

    python_version = config.python_lock_with
    # this check does not make sense on windows as far as I can tell,
    # there is never a "python3.6" or "python2.7" binary just "python"
    if python_version is None and sys.platform != 'win32':
        try:
            # default to configured lock value, then current venv value if it exists, fallback to vulcan's
            # version
            python = get_virtualenv_python()
            python_version = subprocess.check_output(
                [str(python), '-c', 'import sys; print(f"{sys.version_info.major}.{sys.version_info.minor}")'],
                encoding='utf-8').strip()

        except RuntimeError:
            pass
    install_requires, extras_require = asyncio.get_event_loop().run_until_complete(
        resolve_deps_or_report(config, python_version))
    doc = tomlkit.document()
    doc['install_requires'] = tomlkit.array(install_requires).multiline(True)  # type: ignore
    doc['extras_require'] = {
        k: tomlkit.array(v).multiline(True)  # type: ignore
        for k, v in extras_require.items()
    }
    with open(config.lockfile, 'w+') as f:
        f.write(tomlkit.dumps(doc))


@main.command()
@click.argument('req', type=Requirement.parse)
@click.option('--lock/--no-lock', '_lock', default=True)
@pass_vulcan  # order matters, closest to the function definition comes first
@click.pass_context
def add(ctx: click.Context, config: Vulcan, req: Requirement, _lock: bool) -> None:
    "Add new top-level dependency and regenerate lockfile"
    name: str = req.name  # type: ignore
    if req.extras:
        name = f'{name}[{",".join(req.extras)}]'
    try:
        venv_python = get_virtualenv_python()
    except RuntimeError:
        exit("Must be in a virtualenv to use `vulcan add`")
    subprocess.check_call([str(venv_python), '-m', 'pip', 'install', str(req)])
    if req.specifier:  # type: ignore
        # if the user gave a version spec, we blindly take that
        version = str(req.specifier)  # type: ignore
    else:
        # otherwise, we take a freeze to see what was actually installed
        freeze = subprocess.check_output([str(venv_python), '-m', 'pip', 'freeze'], encoding='utf-8').strip()
        try:
            # try and find the thing we just added
            line = next(ln for ln in freeze.split('\n') if ln.startswith(req.name))  # type: ignore
            # and parse it to a version
            spec = packaging.version.parse(str(Requirement.parse(line.strip()).specifier  # type: ignore
                                               )[2:])  # remove the == at the start
            if isinstance(spec, packaging.version.LegacyVersion):
                # this will raise a DeprecationWarning as well, so it will yell at user for us.
                version = ''
            else:
                version = f'~={spec.major}.{spec.minor}'
        except StopIteration:
            # failed to find the thing we just installed, give up.
            version = ''
    with open('pyproject.toml') as f:
        parse = tomlkit.parse(f.read())
    deps = parse['tool']['vulcan'].setdefault('dependencies', tomlkit.table())  # type: ignore
    deps[name] = version
    with open('pyproject.toml', 'w+') as f:
        f.write(tomlkit.dumps(parse))
    if not config.no_lock and _lock:
        ctx.obj = Vulcan.from_source(Path().absolute(), fail_on_missing_lock=False)
        ctx.invoke(lock)


def install_dev_dependencies(target: str = None) -> None:
    config = Vulcan.from_source(Path().absolute(), fail_on_missing_lock=False)

    try:
        virtual_env = get_virtualenv_python()
    except RuntimeError:
        exit('may not use vulcan develop outside of a virtualenv')
    pip_version = get_pip_version(virtual_env)
    if pip_version is None or pip_version < (21, 3):
        print(f"pip version {pip_version} does not support editable installs for PEP517 projects,"
              " Please upgrade your pip")

    if target is not None and target not in config.dev_dependencies:
        raise click.UsageError(f"No such dev dependency {target}")
    if config.dev_dependencies:
        for name, section in config.dev_dependencies.items():
            if target is None or name == target:
                print(f"Installing dev dependencies for {name}", flush=True)
                # print(subprocess.check_output(...)) instead of just subprocess.check_call(...)
                # purely because the CliRunner fixture isn't very good at actually capturing stdout, and it
                # breaks pytest's capsys which _is_ usually good at that
                # ah well
                print(subprocess.check_output(
                    [str(virtual_env), '-m', 'pip', 'install', *(flatten_reqs(section) or [])],
                    encoding='utf-8'),
                      flush=True)


@main.command()
@click.argument('dev_deps_target', required=False, type=str)
def develop(dev_deps_target: Optional[str]) -> None:
    install_develop()
    install_dev_dependencies(target=dev_deps_target)


if __name__ == '__main__':
    main()



import subprocess
import tempfile
import zipfile
from argparse import ArgumentParser
from collections import defaultdict
from configparser import ConfigParser
from io import StringIO
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, cast

import pkginfo  # type: ignore
import tomlkit
from pkg_resources import Requirement


class BuildData(NamedTuple):
    wheel: pkginfo.Wheel
    table: tomlkit.items.Table
    packages: List[str]


def make_parser() -> ArgumentParser:
    parser = ArgumentParser()
    parser.add_argument('--shiv-console-scripts', action='store_true')
    return parser


def wheel() -> BuildData:

    with tempfile.TemporaryDirectory(suffix='.vulcan-migrate') as tmp:
        subprocess.run(['pip', 'wheel', '--no-deps', '-w', tmp, '.'])
        whl = pkginfo.Wheel(next(Path(tmp).glob('*.whl')))
        eps: Dict[str, Dict[str, str]] = defaultdict(dict)
        with zipfile.ZipFile(whl.filename) as zf:
            dist_info = f'{whl.name.replace("-", "_")}-{whl.version}.dist-info'
            try:
                with zf.open(f'{dist_info}/entry_points.txt') as f:
                    cp = ConfigParser()
                    cp.read_file(StringIO(f.read().decode()))
                    for section in cp.sections():
                        for key in cp[section]:
                            eps[section][key] = cp[section][key]
            except KeyError:
                eps = {}
            try:
                with zf.open(f'{dist_info}/top_level.txt') as f:
                    data = f.read().decode()
                    packages = [line.strip() for line in data.split('\n') if line.strip()]
            except KeyError:
                packages = []

    ep_table = tomlkit.table()
    ep_table._is_super_table = True
    for epname, ep in eps.items():
        t = tomlkit.table()
        ep_table[epname] = t
        t.update(ep)
    return BuildData(whl, ep_table, packages)


def contributors(author: Optional[str], author_email: Optional[str]) -> List[tomlkit.items.Table]:
    vals = tomlkit.table()
    if author:
        vals['name'] = author
    if author_email:
        vals['email'] = author_email
    return [vals]


def shiv_from_console_scripts(console_scripts: Dict[str, str]) -> List[Dict[str, str]]:
    shivs: List[Dict[str, str]] = []
    for name in console_scripts:
        shivs.append({
            'bin_name': name,
            'console_script': name,
            'interpreter': '/usr/bin/env python3.6'
        })
    return shivs


def convert() -> None:
    try:
        with open('./pyproject.toml') as f:
            pyproject = tomlkit.loads(f.read())
    except FileNotFoundError:
        pyproject = tomlkit.document()
    if 'project' in pyproject:
        exit('refusing to overwrite current project configuration')
    args = make_parser().parse_args()
    whl, entry_points, packages = wheel()
    project = tomlkit.table()
    vulcan = tomlkit.table()
    pyproject['project'] = project
    if 'tool' not in pyproject:
        pyproject['tool'] = tomlkit.items.Table(
            tomlkit.container.Container(),
            tomlkit.items.Trivia(),
            False,
            is_super_table=True)
    tool = cast(tomlkit.items.Table, pyproject['tool'])
    tool['vulcan'] = vulcan
    project['name'] = whl.name
    project['dynamic'] = ['version']
    if whl.author or whl.author_email:
        project['authors'] = contributors(whl.author, whl.author_email)
    if whl.maintainer or whl.maintainer_email:
        project['maintainers'] = contributors(whl.maintainer, whl.maintainer_email)
    if whl.classifiers:
        project['classifiers'] = tomlkit.array(whl.classifiers).multiline(True)
    if whl.summary:
        project['description'] = whl.summary
    if whl.keywords:
        project['keywords'] = tomlkit.array(whl.keywords.split(',')).multiline(True)
    if whl.license:
        project['license'] = whl.license
    if whl.project_urls:
        project['urls'] = {k: v for k, v in [u.split(', ') for u in whl.project_urls]}
    try:
        readme = next(p for p in Path().iterdir() if p.name.lower().startswith('readme'))
        project['readme'] = str(readme)
    except StopIteration:
        pass

    if whl.requires_dist:
        project['dynamic'].append('dependencies')  # type: ignore[attr-defined]
        vulcan['dependencies'] = {}
        for req in whl.requires_dist:
            parsed_req = Requirement.parse(req)
            if '; extra == "' in str(parsed_req):
                # extra, swing back to this
                continue
            name = parsed_req.name  # type: ignore
            if parsed_req.extras:
                name = f'{name}[{",".join(parsed_req.extras)}]'
            vulcan['dependencies'][name] = str(parsed_req.specifier)  # type: ignore
    if whl.provides_extras:
        project['dynamic'].append('optional-dependencies')  # type: ignore[attr-defined]
        extras = tomlkit.table()
        vulcan['extras'] = extras
        for extra in whl.provides_extras:
            extras[extra] = []
            for req in whl.requires_dist:
                parsed_req = Requirement.parse(req)
                if f'; extra == "{extra}"' not in str(parsed_req):
                    # extra, swing back to this
                    continue
                vulcan['extras'][extra].append(f'{parsed_req.name}{parsed_req.specifier}')  # type: ignore
    if packages:
        vulcan['packages'] = packages
    if whl.requires_python:
        project['requires-python'] = whl.requires_python

    if 'console_scripts' in entry_points:
        project['scripts'] = entry_points['console_scripts']
        del entry_points['console_scripts']
    if 'gui_scripts' in entry_points:
        project['gui-scripts'] = entry_points['gui_scripts']
        del entry_points['gui_scripts']
    if entry_points:
        project['entry-points'] = entry_points
    if args.shiv_console_scripts and 'scripts' in project:
        vulcan['shiv'] = shiv_from_console_scripts(project['scripts'])  # type: ignore

    build_system = tomlkit.table()
    pyproject['build-system'] = build_system
    build_system['requires'] = ['vulcan-py~=1.0']
    build_system['build-backend'] = 'vulcan.build_backend'

    with open('./pyproject.toml', 'w+') as f:
        f.write(tomlkit.dumps(pyproject))



import asyncio
import tempfile
from itertools import chain
from typing import Dict, List, Tuple

from pkg_resources import Requirement

from vulcan.isolation import VulcanEnvBuilder, create_venv


async def build_requires(pipenv: VulcanEnvBuilder, requires: List[str]) -> Dict[Requirement, Requirement]:
    with tempfile.TemporaryDirectory() as site_packages:
        await pipenv.install(site_packages, requires)
        freeze = await pipenv.freeze(site_packages)
        return freeze


async def resolve_deps(install_requires: List[str], extras: Dict[str, List[str]],
                       python_version: str = None
                       ) -> Tuple[List[str], Dict[str, List[str]]]:

    if not install_requires and not extras:
        return [], {}

    extras_list = list(extras.items())
    with create_venv(python_version) as pipenv:
        print("Building base requires")
        base_freeze = await build_requires(pipenv, install_requires)
        if not extras_list:
            # if we have no extras, we are done here.
            return sorted([str(req) for req in base_freeze.values()]), {}

        print("Building requirements for base + all extras")
        final_out_task = asyncio.get_event_loop().create_task(
            build_requires(
                pipenv,
                install_requires + list(chain.from_iterable(reqs for _, reqs in extras_list))))
        resolved_extras = {}
        for extra, extra_reqs in extras_list:
            # this is the expensive bit, because we create a new venv for each extra
            print(f"Building requirements for extra '{extra}'")
            resolved_extras[extra] = asyncio.get_event_loop().create_task(
                build_requires(pipenv, install_requires + extra_reqs))

        # It is important here to wait until ALL tasks are complete (return_exceptions=True) because on
        # windows, if that is not done then the TemporaryDirectory is create_venv will try to remove itself
        # while our async subprocesses still have a lock on the python.exe (windows-only issue). which then
        # causes more errors
        results = await asyncio.gather(*resolved_extras.values(), final_out_task, return_exceptions=True)
        for res in results:
            if isinstance(res, BaseException):
                raise res
        all_resolved = final_out_task.result()

        extras_out = {k: sorted([str(all_resolved[req]) for req in v.result()])
                      for k, v in resolved_extras.items()}
        return sorted([str(all_resolved[req]) for req in base_freeze.keys()]), extras_out



from pathlib import Path
import sys

import tomlkit


def convert() -> None:
    with open('./pyproject.toml') as f:
        pyproject = tomlkit.loads(f.read())
        project = pyproject.get('project')
        if not project:
            exit("Refusing to try and migrate a pyproject.toml without any [project] section")
        tool = pyproject.setdefault('tool', tomlkit.table(is_super_table=True))
        setuptools = tool.setdefault('setuptools', tomlkit.table(is_super_table=True))
        vulcan = tool.get('vulcan', {})
        setuptools_dynamic = setuptools.setdefault('dynamic', tomlkit.table(is_super_table=True))
    dynamic = project.get('dynamic', [])

    # fix dynamic
    for vulcan_key, real_key in (('dependencies', 'dependencies'),
                                 ('version', 'version'), ('extras', 'optional-dependencies')):
        if vulcan_key in vulcan and ('dynamic' not in project or real_key not in project['dynamic']):
            dynamic.append(real_key)

    # fix version discovery
    version_file = next(Path().rglob('VERSION'), None)
    if 'version' not in project and version_file is not None:
        dynamic.append('version')
        if 'version' in setuptools_dynamic:
            print("version already configured in [tool.setuptools.dynamic], not changing", file=sys.stderr)
        else:
            version_table = tomlkit.table(is_super_table=True)
            setuptools_dynamic['version'] = version_table
            version_table['file'] = str(version_file)
        if 'version' in vulcan:
            del vulcan['version']

    # fix packages
    if 'packages' in vulcan:
        setuptools_packages = setuptools.setdefault('packages', tomlkit.table(is_super_table=True))
        find = setuptools_packages.setdefault('find', tomlkit.table(is_super_table=True))
        include = find.setdefault('include', tomlkit.array())
        include.extend(vulcan['packages'])
        del vulcan['packages']

    build_system = tomlkit.table()
    pyproject['build-system'] = build_system
    build_system['requires'] = ['vulcan-py~=2.0']
    build_system['build-backend'] = 'vulcan.build_backend'

    if dynamic:
        project['dynamic'] = dynamic

    with open('./pyproject.toml', 'w+') as f:
        f.write(tomlkit.dumps(pyproject))



import os
import re
import shutil
import subprocess
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, Dict, Generator, List, Optional, Tuple

from editables import EditableProject  # type: ignore

from vulcan import Vulcan
from vulcan.plugins import PluginRunner

version: Callable[[str], str]
if sys.version_info >= (3, 8):
    from importlib.metadata import version
else:
    from importlib_metadata import version

__all__ = ['build_wheel', 'build_sdist']


@contextmanager
def patch_argv(argv: List[str]) -> Generator[None, None, None]:
    old_argv = sys.argv[:]
    sys.argv = [sys.argv[0]] + argv
    yield
    sys.argv = old_argv


def build(outdir: str, config_settings: Dict[str, str] = None) -> str:
    config = Vulcan.from_source(Path().absolute())

    # https://setuptools.readthedocs.io/en/latest/userguide/keywords.html
    # https://docs.python.org/3/distutils/apiref.html
    with PluginRunner(config):
        dist = config.setup()
    rel_dist = Path(dist.dist_files[0][-1])  # type: ignore[attr-defined]
    shutil.move(str(rel_dist), Path(outdir) / rel_dist.name)
    return rel_dist.name


def build_wheel(wheel_directory: str, config_settings: Dict[str, str] = None, metadata_directory: str = None) -> str:
    with patch_argv(['bdist_wheel']):
        return build(wheel_directory, config_settings)


def build_sdist(
    sdist_directory: str,
    config_settings: Dict[str, str] = None,
) -> str:
    with patch_argv(['sdist']):
        return build(sdist_directory, config_settings)


def get_virtualenv_python() -> Path:
    virtual_env = os.environ.get('VIRTUAL_ENV')
    if virtual_env is None:
        raise RuntimeError("No virtualenv active")
    if sys.platform == 'win32':
        # sigh
        return Path(virtual_env, 'Scripts', 'python')
    else:
        # if this isn't in an else,
        # mypy complains on windows that it is unreachable
        return Path(virtual_env, 'bin', 'python')


# tox requires these two for some reason :(
def get_requires_for_build_sdist(config_settings: Dict[str, str] = None) -> List[str]:
    return []


def get_requires_for_build_wheel(config_settings: Dict[str, str] = None) -> List[str]:
    return []


def get_pip_version(python_callable: Path) -> Optional[Tuple[int, ...]]:
    out = subprocess.check_output([str(python_callable), '-m', 'pip', '--version'], encoding='utf-8')
    m = re.search(r'pip (\d+\.\d+(\.\d+)?)', out)
    if not m:
        return None
    return tuple((int(n) for n in m.group(1).split('.')))


def install_develop() -> None:
    config = Vulcan.from_source(Path().absolute())

    try:
        virtual_env = get_virtualenv_python()
    except RuntimeError:
        exit('may not use vulcan develop outside of a virtualenv')
    pip_version = get_pip_version(virtual_env)
    if pip_version is None or pip_version < (21, 3):
        print(f"pip version {pip_version} does not support editable installs for PEP517 projects,"
              " Please upgrade your pip")

    path = str(Path().absolute())
    if config.configured_extras:
        path = f'{path}[{",".join(config.configured_extras)}]'
    subprocess.check_call([str(virtual_env), '-m', 'pip', 'install', '-e', path])


# pep660 functions
def unpack(whl: Path) -> Path:
    with tempfile.TemporaryDirectory() as tmp:
        subprocess.check_output(f'wheel unpack {whl} -d {tmp}'.split())
        unpacked = list(Path(tmp).glob('*'))
        assert len(unpacked) == 1
        shutil.copytree(unpacked[0], whl.parent / unpacked[0].name)
        return whl.parent / unpacked[0].name


def pack(unpacked_wheel: Path) -> Path:
    with tempfile.TemporaryDirectory() as tmp:
        subprocess.check_output(f'wheel pack {unpacked_wheel} -d {tmp}'.split())
        packed = list(Path(tmp).glob('*.whl'))
        assert len(packed) == 1
        shutil.copy(packed[0], unpacked_wheel.parent)
        return unpacked_wheel.parent / packed[0].name


def add_requirement(unpacked_whl_dir: Path, req: str) -> None:
    metadata = next(unpacked_whl_dir.glob('*.dist-info')) / 'METADATA'  # is mandatory
    with metadata.open() as f:
        metadata_lines = list(f)
    i = 0
    for i, line in enumerate(metadata_lines):
        if not (line.strip() and not line.startswith('Requires-Dist: ')):
            # find the start of the requires-dist, or the end of the metadata keys
            break
    metadata_lines.insert(i, f'Requires-Dist: {req}\n')
    metadata.write_text(''.join(metadata_lines))


def make_editable(whl: Path) -> None:
    unpacked_whl_dir = unpack(whl)
    add_requirement(unpacked_whl_dir, f"editables (~={version('editables')})")
    # https://www.python.org/dev/peps/pep-0427/#escaping-and-unicode
    # this is guarenteed to exist, name is extremely mandatory. Can't make a valid wheel without it.
    # it might be UNKNOWN, but this is user error.
    name = next(
        line.split(':')[1].strip()
        for line in (next(unpacked_whl_dir.glob('*.dist-info')) / 'METADATA').read_text().splitlines()
        if 'Name:' in line)
    project_name = re.sub(r'[^\w\d.]+', '_', name, re.UNICODE)
    project = EditableProject(project_name, Path().absolute())
    packages = (p for p in unpacked_whl_dir.iterdir() if not p.name.endswith('.dist-info'))
    for package in packages:
        project.map(package.name, package.name)
        # removing the actual code packages because they will conflict with the .pth files, and take
        # precendence over them
        shutil.rmtree(unpacked_whl_dir / package.name)
    for name, content in project.files():
        (unpacked_whl_dir / name).write_text(content)

    assert whl == pack(unpacked_whl_dir), 'pre-wheel and post-wheel should be the same path'
    shutil.rmtree(unpacked_whl_dir)


def build_editable(wheel_directory: str, config_settings: Dict[str, str] = None, metadata_directory: str = None) -> str:
    whl_path = Path(wheel_directory) / build_wheel(wheel_directory, config_settings, metadata_directory)
    make_editable(whl_path)
    return whl_path.name



import distutils.core
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union, cast

import tomlkit
import tomlkit.container
import tomlkit.items
from setuptools import setup
from typing_extensions import TypedDict


class VulcanConfigError(Exception):
    pass


class VersionDict(TypedDict, total=False):
    version: str
    extras: List[str]


VersionSpecs = Mapping[str, Union[str, VersionDict]]


def flatten_reqs(versions: VersionSpecs) -> List[str]:
    return [to_pep508(lib, req) for lib, req in versions.items()]


@dataclass
class ShivOpts:
    bin_name: str
    console_script: Optional[str] = None
    entry_point: Optional[str] = None
    interpreter: Optional[str] = None
    with_extras: Optional[List[str]] = None
    extra_args: str = ''


def list_or_none(val: Any) -> Optional[List[str]]:
    return [str(v) for v in val] if val is not None else None


def str_or_none(val: Any) -> Optional[str]:
    return str(val) if val is not None else None


def dict_or_none(val: Any) -> Optional[Dict[str, Any]]:
    return {str(k): v for k, v in val.items()} if val is not None else None


@dataclass
class Vulcan:
    source_path: Path
    plugins: Optional[List[str]]
    shiv_options: List[ShivOpts]
    lockfile: Path
    dependencies: Optional[List[str]]
    configured_dependencies: VersionSpecs
    extras: Optional[Dict[str, List[str]]]
    configured_extras: Dict[str, List[str]]
    dev_dependencies: Dict[str, VersionSpecs]
    dynamic: Optional[List[str]]
    no_lock: bool = False
    python_lock_with: Optional[str] = None

    @classmethod
    def from_source(cls, source_path: Path, fail_on_missing_lock: bool = True) -> 'Vulcan':
        with open(source_path / 'pyproject.toml') as f:
            all_config = tomlkit.loads(f.read())
            config = all_config['tool']['vulcan']  # type: ignore[index]
            assert isinstance(config, dict)
            dynamic = all_config['project'].get('dynamic', [])  # type: ignore[union-attr]
        lockfile = source_path / config.get('lockfile', 'vulcan.lock')

        no_lock = config.get('no-lock', False)
        install_requires: Optional[List[str]] = []
        extras_require: Optional[Dict[str, List[str]]] = {}
        if not no_lock:
            try:
                install_requires, extras_require = get_requires(lockfile)
            except FileNotFoundError:
                if fail_on_missing_lock:
                    raise
                install_requires = None
                extras_require = None

        python_lock_with = config.get('python-lock-with')

        shiv_ops = []
        shiv_config = config.get('shiv', [])
        for conf in shiv_config:
            shiv_ops.append(
                ShivOpts(
                    bin_name=str(conf.get('bin_name')),
                    console_script=str_or_none(conf.get('console_script')),
                    entry_point=str_or_none(conf.get('entry_point')),
                    interpreter=str_or_none(conf.get('interpreter')),
                    with_extras=[str(e) for e in conf.get('with_extras', [])],
                    extra_args=str(conf.get('extra_args', '')),
                ))
        # note that setuptools also checks this, and says that it _should_ consider this a warning, but will
        # not for the intermediate period.
        # We'll consider it an error until setuptools sees fit to do it for us, and then remove this check.
        if 'dependencies' in config and 'dependencies' not in dynamic:
            raise RuntimeError("tool.vulcan.dependencies configured but 'dependencies' not in dynamic,"
                               " this is an error according to PEP-621. "
                               "See https://peps.python.org/pep-0621/#dynamic for more information")
        if 'extras' in config and 'optional-dependencies' not in dynamic:
            raise RuntimeError("tool.vulcan.extras configured but 'optional-dependencies' not in dynamic,"
                               " this is an error according to PEP-621. "
                               "See https://peps.python.org/pep-0621/#dynamic for more information")

        return cls(source_path=source_path,
                   plugins=list_or_none(config.get('plugins')),
                   lockfile=lockfile,
                   shiv_options=shiv_ops,
                   dependencies=install_requires,
                   configured_dependencies=config.get('dependencies', {}),
                   extras=extras_require,
                   dev_dependencies=config.get('dev-dependencies', {}),
                   configured_extras=config.get('extras', {}),
                   no_lock=no_lock,
                   python_lock_with=python_lock_with,
                   dynamic=dynamic)

    def setup(self, config_settings: Dict[str, str] = None) -> distutils.core.Distribution:
        install_requires: Optional[List[str]]
        extras_require: Optional[Dict[str, List[str]]]
        if self.no_lock or (config_settings and config_settings.get('no-lock') == 'true'):
            install_requires = flatten_reqs(self.configured_dependencies)
            extras_require = self.configured_extras
        else:
            install_requires = self.dependencies
            extras_require = self.extras
        # setuptools apparently does not know what setuptools returns
        # very reassuring
        return setup(  # type: ignore[no-any-return,func-returns-value]
            install_requires=install_requires,
            extras_require=extras_require)


def get_requires(lockfile: Path) -> Tuple[List[str], Dict[str, List[str]]]:
    if not lockfile.exists():
        raise FileNotFoundError(f"Expected lockfile {lockfile}, does not exist")
    with lockfile.open() as f:
        content = cast(tomlkit.container.Container, tomlkit.loads(f.read()))
    return (
        list(content['install_requires']),  # type: ignore
        {k: list(v)
         for k, v in content['extras_require'].items()})  # type: ignore


def to_pep508(lib: str, req: Union[str, VersionDict]) -> str:
    if not isinstance(req, (str, dict)):
        raise VulcanConfigError(f"Invalid requirement {req} -- must be a dict or a string")
    if isinstance(req, str):
        # e.g. "example_lib", "~=1.2.3" -> "example_lib~=1.2.3"
        return f'{lib}{req}'
    try:
        extras = f'[{",".join(req["extras"])}]' if 'extras' in req else ''
        # "example_lib", {"version": "~=1.2.3", "extras"=["networkx","git"]}
        #                           -> "example_lib[networkx,git]~=1.2.3"
        # "example_lib", {"version": "~=1.2.3"} -> "example_lib~=1.2.3"
        return f'{lib}{extras}{req["version"]}'
    except KeyError as e:
        raise VulcanConfigError(f'invalid requirement {lib} ({req}) -- {e}') from e
